/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 *  GNOME Print Manager
 *  Copyright (C) 2002 Sun Microsystems, Inc.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <config.h>

#include "print-manager-queue-solaris.h"

#include "lpstat-helper.h"
#include "print-prefs.h"

#include <string.h>
#include <stdlib.h>
#include <gnome.h>

#define PARENT_TYPE (PRINT_MANAGER_TYPE_QUEUE)

/* Static functions */

struct _PrintManagerQueueSolarisPrivate {
	gchar       *queue_name;
	gboolean     available;
	gchar       *status;
	gchar       *device_name;
	gchar       *device_status;
	gchar       *host_name;
	gchar       *host_queue;

	gboolean     new_available;
	gchar       *new_status;
	gchar       *new_device_name;
	gchar       *new_device_status;
	gchar       *new_host_name;
	gchar       *new_host_queue;

	gint         steps;

	gint         count;
	PrintJob   **jobs;

	gint         new_count;
	PrintJob   **new_jobs;


	guint	     is_remote : 1;
	guint	     new_is_remote : 1;

	guint        in_reload : 1;
	guint        is_loaded : 1;
	int          reload_cnt;
};

static void
merge_jobs (PrintManagerQueueSolaris *queue)
{
	int new_count = queue->priv->new_count;
	int count = queue->priv->count;
	int i;
	int j;
	for (i = 0; i < new_count; i++) {
		for (j = 0; j < count; j++) {
			if (queue->priv->jobs[j] &&
			    print_job_equal (queue->priv->new_jobs[i],
					     queue->priv->jobs[j])) {
				print_job_free (queue->priv->new_jobs[i]);
				queue->priv->new_jobs[i] = queue->priv->jobs[j];
				queue->priv->jobs[j] = NULL;
				break;
			}
		}
	}
	for (j = 0; j < count; j++) {
		if (queue->priv->jobs[j])
			print_job_free (queue->priv->jobs[j]);
	}
	g_free (queue->priv->jobs);
	queue->priv->jobs = queue->priv->new_jobs;
	queue->priv->count = queue->priv->new_count;
	queue->priv->new_jobs = NULL;
	queue->priv->new_count = 0;
}

static void
one_step_finished (PrintManagerQueueSolaris *queue)
{
	queue->priv->steps --;
	if (queue->priv->steps == 0) {
		g_free (queue->priv->status);
		g_free (queue->priv->device_name);
		g_free (queue->priv->device_status);
		g_free (queue->priv->host_name);
		g_free (queue->priv->host_queue);

		queue->priv->status        = queue->priv->new_status;
		queue->priv->available     = queue->priv->new_available;
		queue->priv->device_name   = queue->priv->new_device_name;
		queue->priv->device_status = queue->priv->new_device_status;
		queue->priv->host_name     = queue->priv->new_host_name;
		queue->priv->host_queue    = queue->priv->new_host_queue;

		queue->priv->is_remote     = queue->priv->new_is_remote;

		queue->priv->new_status        = NULL;
		queue->priv->new_available     = TRUE;
		queue->priv->new_device_name   = NULL;
		queue->priv->new_device_status = NULL;
		queue->priv->new_host_name     = NULL;
		queue->priv->new_host_queue    = NULL;

		queue->priv->new_is_remote     = FALSE;

		merge_jobs (queue);

		queue->priv->is_loaded = TRUE;

		print_manager_queue_changed (PRINT_MANAGER_QUEUE (queue));
		print_manager_queue_changes_done (PRINT_MANAGER_QUEUE (queue));

		queue->priv->in_reload = FALSE;
		if (queue->priv->reload_cnt > 0) {
			queue->priv->reload_cnt--;
			print_manager_queue_reload (
				PRINT_MANAGER_QUEUE (queue));
		}
	}
}

static void 
list_jobs_output (gint       lpstat_id,
		  gchar     *line,
		  gpointer   user_data)
{
	PrintManagerQueueSolaris *queue = user_data;
	PrintJob *job;
	gchar *jsize;

	if (!strchr (line, '|'))
		return;

	job = g_new0 (PrintJob, 1);	
	job->queue_name  = g_strdup (queue->priv->queue_name);

	strtok (line, "|"); /* printer */
	job->doc_name    = g_strdup (strtok (NULL, "|"));
	job->job_num     = g_strdup (strtok (NULL, "|"));
	job->owner       = g_strdup (strtok (NULL, "|"));
	job->user_owned  = !strcmp (job->owner, g_get_user_name ());
	job->submit_date = g_strdup (strtok (NULL, "|"));
	job->status      = PRINT_JOB_STATUS_NORMAL;

	jsize = strtok (NULL, "\n");
	if (jsize && strlen (jsize))
		job->file_size = g_strdup_printf (_("%s bytes"), jsize);
	else 
		job->file_size = g_strdup (_("Unknown"));

	queue->priv->new_jobs = g_renew (PrintJob *,
					 queue->priv->new_jobs,
					 queue->priv->new_count + 1);

	queue->priv->new_jobs [queue->priv->new_count] = job;
	queue->priv->new_count++;
}

static void
list_jobs_end (gint       lpstat_id,
	       gboolean   cancelled,
	       gpointer   user_data)
{
	PrintManagerQueueSolaris *queue = user_data;

	one_step_finished (queue);
}

static void
list_jobs (PrintManagerQueueSolaris *queue)
{
	if (queue->priv->new_is_remote) {
		gchar *command_args [] = { queue->priv->new_host_name, 
					   queue->priv->new_host_queue, 
					   NULL };
		lpstat_run (list_jobs_output,
			    NULL,
			    list_jobs_end,
			    queue,
			    LPSTAT_LIST_REMOTE_JOBS,
			    command_args);
	} else {
		gchar *command_args [] = { queue->priv->queue_name, NULL };
		lpstat_run (list_jobs_output,
			    NULL,
			    list_jobs_end,
			    queue,
			    LPSTAT_LIST_LOCAL_JOBS,
			    command_args);
	}
}

static void
get_queue_status_output (gint       lpstat_id,
			 gchar     *line,
			 gpointer   user_data)
{
	PrintManagerQueueSolaris *queue = user_data;

	if (strlen (line)) {
		queue->priv->new_status = g_strdup (_("Not Active"));
		queue->priv->new_available = FALSE;
	}
}

static void
get_device_status_output (gint       lpstat_id,
			  gchar     *line,
			  gpointer   user_data)
{
	PrintManagerQueueSolaris *queue = user_data;

	if (strlen (line)) {
		queue->priv->new_device_status = g_strdup (_("Not Active"));
		queue->priv->new_available = FALSE;
	}
}

static void
get_queue_or_device_status_end (gint       lpstat_id,
				gboolean   cancelled,
				gpointer   user_data)
{
	PrintManagerQueueSolaris *queue = user_data;

	one_step_finished (queue);
}

static void
get_queue_status (PrintManagerQueueSolaris *queue)
{
	gchar *command_args[] = { queue->priv->queue_name, NULL };

	lpstat_run (get_queue_status_output,
		    NULL,
		    get_queue_or_device_status_end,
		    queue,
		    LPSTAT_GET_QUEUE_STATUS,
		    command_args);

	queue->priv->new_status = g_strdup (_("Active"));

	if (!queue->priv->new_is_remote) {
		lpstat_run (get_device_status_output, 
			    NULL, 
			    get_queue_or_device_status_end,
			    queue, 
			    LPSTAT_GET_DEVICE_STATUS,
			    command_args);
		queue->priv->new_device_status = g_strdup (_("Active"));
	}
}

static void 
get_queue_output (gint       lpstat_id,
		  gchar     *line,
		  gpointer   user_data)
{
	PrintManagerQueueSolaris *queue = user_data;
	gchar **split;

	split = g_strsplit (line, ":", 0);
	if (!split || !split [0])
		return;

	if (split [0] && strlen (split [0])) {
		queue->priv->new_device_name = g_strdup (split [0]);
		queue->priv->new_is_remote = FALSE;
	} 
	else if (split [1] && strlen (split [1])) {
		queue->priv->new_host_name = g_strdup (split [1]);
		queue->priv->new_host_queue = 
			g_strdup (strlen (split [2]) ? 
				  	split [2] : 
				  	queue->priv->queue_name);

		queue->priv->new_is_remote = TRUE;
	}

	g_strfreev (split);
}

static void
get_queue_end (gint       lpstat_id,
	       gboolean   cancelled,
	       gpointer   user_data)
{
	PrintManagerQueueSolaris *queue = user_data;

	if (queue->priv->new_is_remote)
		queue->priv->steps = 2;
	else
		queue->priv->steps = 3;

	get_queue_status (queue);
	list_jobs (queue);
}

typedef struct {
	PrintManagerQueue  *queue;
	PrintJob           *job;
	PrintJobCanceledFn  callback;
	gpointer           *user_data;
	gchar              *errstr;
} JobCancelData;

static void
job_cancel_error_output (gint      lpstat_id,
			 gchar    *line,
			 gpointer  user_data)
{
	JobCancelData *data = user_data;

	if (data->errstr)
		data->errstr = g_strconcat (data->errstr, line, NULL);
	else
		data->errstr = g_strdup (line);
}

static void
job_cancel_output (gint      lpstat_id,
		   gchar    *line,
		   gpointer  user_data)
{
	gchar *colon;

	colon = strrchr (line, ':');
	if (colon)
		line = colon + 1;

	if (strstr (line, "cancelled"))
		return;

	job_cancel_error_output (lpstat_id, line, user_data);
}

static void
job_cancel_end (gint      lpstat_id,
		gboolean  cancelled,
		gpointer  user_data)
{
	JobCancelData *data = user_data;
	PrintManagerQueueSolaris *pmqs;
	gboolean success = data->errstr ? FALSE : TRUE;

	pmqs = PRINT_MANAGER_QUEUE_SOLARIS (data->queue);

	if (data->callback)
		(*data->callback) (data->queue,
				   data->job, 
				   success,
				   data->errstr, 
				   data->user_data);

	if (success) {
		PrintManagerQueueSolaris *pmqs = PRINT_MANAGER_QUEUE_SOLARIS (data->queue);
		int count = pmqs->priv->count;
		int i;

		for (i = 0; i < count; i++) {
			if (print_job_equal (data->job, pmqs->priv->jobs[i])) {
				print_job_free (pmqs->priv->jobs[i]);
				memmove (pmqs->priv->jobs + i, pmqs->priv->jobs + i + 1, (count - i - 1) * sizeof (PrintJob *));
				pmqs->priv->count --;
				break;
			}
		}
	} else {
		data->job->status = PRINT_JOB_STATUS_NORMAL;
	}

	if (pmqs->priv->is_loaded) {
		print_manager_queue_changed (data->queue);
		print_manager_queue_changes_done (data->queue);
	}

	g_free (data->errstr);
	g_free (data);
}

/* access functions */
static gint
pmqs_get_count (PrintManagerQueue *pmq)
{
	PrintManagerQueueSolaris *pmqs = PRINT_MANAGER_QUEUE_SOLARIS (pmq);

	g_return_val_if_fail (pmqs != NULL, -1);

	return pmqs->priv->count;
}

static PrintJob *
pmqs_get_job (PrintManagerQueue *pmq, int n)
{
	PrintManagerQueueSolaris *pmqs = PRINT_MANAGER_QUEUE_SOLARIS (pmq);

	g_return_val_if_fail (pmqs != NULL, NULL);
	g_return_val_if_fail (n >= 0, NULL);
	g_return_val_if_fail (n < pmqs->priv->count, NULL);

	return pmqs->priv->jobs[n];
}


static const gchar *
pmqs_get_queue_name (PrintManagerQueue *pmq)
{
	PrintManagerQueueSolaris *pmqs = PRINT_MANAGER_QUEUE_SOLARIS (pmq);

	return pmqs->priv->queue_name;
}

static const gchar *
pmqs_get_status (PrintManagerQueue *pmq)
{
	PrintManagerQueueSolaris *pmqs = PRINT_MANAGER_QUEUE_SOLARIS (pmq);

	return pmqs->priv->status;
}

static gboolean
pmqs_get_available (PrintManagerQueue *pmq)
{
	PrintManagerQueueSolaris *pmqs = PRINT_MANAGER_QUEUE_SOLARIS (pmq);

	return pmqs->priv->available;
}

static const gchar *
pmqs_get_device_name (PrintManagerQueue *pmq)
{
	PrintManagerQueueSolaris *pmqs = PRINT_MANAGER_QUEUE_SOLARIS (pmq);

	return pmqs->priv->device_name;
}

static const gchar *
pmqs_get_device_status (PrintManagerQueue *pmq)
{
	PrintManagerQueueSolaris *pmqs = PRINT_MANAGER_QUEUE_SOLARIS (pmq);

	return pmqs->priv->device_status;
}

static const gchar *
pmqs_get_host_name (PrintManagerQueue *pmq)
{
	PrintManagerQueueSolaris *pmqs = PRINT_MANAGER_QUEUE_SOLARIS (pmq);

	return pmqs->priv->host_name;
}

static const gchar *
pmqs_get_host_queue (PrintManagerQueue *pmq)
{
	PrintManagerQueueSolaris *pmqs = PRINT_MANAGER_QUEUE_SOLARIS (pmq);

	return pmqs->priv->host_queue;
}

static gboolean
pmqs_get_is_remote (PrintManagerQueue *pmq)
{
	PrintManagerQueueSolaris *pmqs = PRINT_MANAGER_QUEUE_SOLARIS (pmq);

	return pmqs->priv->is_remote;
}

static void
pmqs_cancel_job (PrintManagerQueue  *pmq,
		 PrintJob           *job,
		 PrintJobCanceledFn  callback,
		 gpointer            user_data)
{
	JobCancelData *data;

	g_return_if_fail (job != NULL);

	if (job->status == PRINT_JOB_STATUS_CANCELLING) {
		if (callback)
			(*callback) (pmq,
				     job,
				     FALSE,
				     _("Already Cancelling"),
				     user_data);
		return;
	}

	data = g_new0 (JobCancelData, 1);
	data->queue = pmq;
	data->job = job;
	data->callback = callback;
	data->user_data = user_data;

	{ 
		gchar *command_args[] = { job->queue_name, job->job_num };

		lpstat_run (job_cancel_output,
			    job_cancel_error_output,
			    job_cancel_end,
			    data,
			    LPSTAT_CANCEL_JOB,
			    command_args);

		job->status = PRINT_JOB_STATUS_CANCELLING;
		if (PRINT_MANAGER_QUEUE_SOLARIS (pmq)->priv->is_loaded) {
			print_manager_queue_changed (pmq);
			print_manager_queue_changes_done (pmq);
		}
	}
}

typedef struct {
	GSList             *list;
	GSList             *current;
	PrintJobCanceledFn  callback;
	gpointer            user_data;
} JobCancelListData;

static void
cancel_jobs_internal (PrintManagerQueue *queue,
		      PrintJob    *job,
		      gboolean     success,
		      gchar       *errstr,
		      gpointer     user_data)
{
	JobCancelListData *data = user_data;

	(*data->callback) (queue, job, success, errstr, data->user_data);

	data->current = data->current->next;

	if (!data->current) {
		g_slist_free (data->list);
		g_free (data);
	} 
	else {
		PrintJob *next_job = data->current->data;

		pmqs_cancel_job (queue,
				 next_job,
				 cancel_jobs_internal,
				 data);
	}
}

static void
pmqs_cancel_jobs (PrintManagerQueue  *pmq,
		  GSList             *jobs,
		  PrintJobCanceledFn  callback,
		  gpointer            user_data)
{
	JobCancelListData *data;
	PrintJob *first_job;

	g_return_if_fail (jobs != NULL);

	data = g_new0 (JobCancelListData, 1);
	data->list      = g_slist_copy (jobs);
	data->current   = data->list;
	data->callback  = callback;
	data->user_data = user_data;

	first_job = data->current->data;

	pmqs_cancel_job (pmq, 
			 first_job, 
			 cancel_jobs_internal, 
			 data);
}

/* Change functions */
static void
pmqs_reload (PrintManagerQueue *pmq)
{
	PrintManagerQueueSolaris *pmqs = PRINT_MANAGER_QUEUE_SOLARIS (pmq);

	if (!pmqs->priv->in_reload) {
		gchar *command_args[] = { pmqs->priv->queue_name, NULL };

		pmqs->priv->new_count = 0;
		pmqs->priv->new_jobs = NULL;

		lpstat_run (get_queue_output,
			    NULL,
			    get_queue_end,
			    pmqs,
			    LPSTAT_GET_QUEUE_ATTRS,
			    command_args);
		pmqs->priv->in_reload = TRUE;
	} else
		pmqs->priv->reload_cnt++;
}

static gboolean
pmqs_get_is_loaded (PrintManagerQueue *pmq)
{
	PrintManagerQueueSolaris *pmqs = PRINT_MANAGER_QUEUE_SOLARIS (pmq);

	return pmqs->priv->is_loaded;
}

static void
pmqs_class_init (PrintManagerQueueSolarisClass *klass)
{
	PrintManagerQueueClass *queue_class = PRINT_MANAGER_QUEUE_CLASS (klass);

	queue_class->get_count         = pmqs_get_count;
	queue_class->get_job           = pmqs_get_job;

	queue_class->get_queue_name    = pmqs_get_queue_name;
	queue_class->get_status        = pmqs_get_status;
	queue_class->get_available     = pmqs_get_available;
	queue_class->get_device_name   = pmqs_get_device_name;
	queue_class->get_device_status = pmqs_get_device_status;
	queue_class->get_host_name     = pmqs_get_host_name;
	queue_class->get_host_queue    = pmqs_get_host_queue;
	queue_class->get_is_remote     = pmqs_get_is_remote;

	queue_class->cancel_job        = pmqs_cancel_job;
	queue_class->cancel_jobs       = pmqs_cancel_jobs;

	queue_class->reload            = pmqs_reload;
	queue_class->get_is_loaded     = pmqs_get_is_loaded;
}

static void
pmqs_init (PrintManagerQueueSolaris *queue)
{
	queue->priv = g_new (PrintManagerQueueSolarisPrivate, 1);
 
	queue->priv->queue_name        = NULL;
	queue->priv->available         = TRUE;
	queue->priv->status            = NULL;
	queue->priv->device_name       = NULL;
	queue->priv->device_status     = NULL;
	queue->priv->host_name         = NULL;
	queue->priv->host_queue        = NULL;

	queue->priv->is_remote         = FALSE;
	
	queue->priv->new_available     = TRUE;
	queue->priv->new_status        = NULL;
	queue->priv->new_device_name   = NULL;
	queue->priv->new_device_status = NULL;
	queue->priv->new_host_name     = NULL;
	queue->priv->new_host_queue    = NULL;

	queue->priv->new_is_remote     = FALSE;

	queue->priv->count             = 0;
	queue->priv->jobs              = NULL;

	queue->priv->new_count         = 0;
	queue->priv->new_jobs          = NULL;

	queue->priv->in_reload         = FALSE;
	queue->priv->is_loaded         = FALSE;
	queue->priv->reload_cnt        = 0;
}

/**
 * print_manager_queue_solaris_get_type:
 * @void: 
 * 
 * Registers the &PrintManagerQueueSolaris class if necessary, and returns
 * the type ID associated to it.
 * 
 * Return value: The type ID of the &PrintManagerQueueSolaris class.
 **/
GType
print_manager_queue_solaris_get_type (void)
{
	static GType type = 0;

	if (!type) {
		GTypeInfo info = {
			sizeof (PrintManagerQueueSolarisClass),
			(GBaseInitFunc) NULL,
			(GBaseFinalizeFunc) NULL,
			(GClassInitFunc) pmqs_class_init,
			(GClassFinalizeFunc) NULL,
			NULL, /* class_data */
			sizeof (PrintManagerQueueSolaris),
			0,
			(GInstanceInitFunc) pmqs_init,
			NULL /* value_table */
		};

		type = g_type_register_static (PARENT_TYPE,
					       "PrintManagerQueueSolaris",
					       &info, 0);
	}

	return type;
}

PrintManagerQueue *
print_manager_queue_solaris_new (char *name)
{
	PrintManagerQueueSolaris *queue = 
		PRINT_MANAGER_QUEUE_SOLARIS (
			g_type_create_instance (
				PRINT_MANAGER_TYPE_QUEUE_SOLARIS));

	queue->priv->queue_name = g_strdup (name);

	return PRINT_MANAGER_QUEUE (queue);
}
